
namespace CSLint
{
    using System;
    using System.Reflection;
    using System.Collections;
    using System.Threading;
    using System.Reflection.Emit;
    using Reflector.Disassembler;
    using System.Runtime.CompilerServices;

    public class CSLint
    {
        public static void Main(String[] args)
        {
            CSLint.instance.proceed(args);
        }

        internal void proceed(String[] args)
        {
            load(args);
            build();
            verify();
        }


        internal void build()
        {
            foreach (ClassDesc desc in classes.Values) 
            {
                desc.addOverriddenMethodInvocations();
            }
            for (MethodDesc entry = entryPoints; entry != null; entry = entry.nextEntryPoint)
            {
                entry.buildConcurrentClosure();
            }
            if (allGlobalsReentrant) 
            {
                foreach (Module module in modules) 
                {
                    foreach (MethodBase mb in module.GetMethods()) 
                    {
                        MethodDesc md = getMethodDesc(mb);
                        if (md != null) 
                        { 
                            md.buildConcurrentClosure();
                        }
                    }
                }
            }
        }

        internal void verify()
        {
            GraphVertex.verify();
        }


        internal void loadMethodBody(MethodBase methodBase) 
        {   
            if (methodBase != null) 
            {
                MethodBody methodBody = reader.GetMethodBody(methodBase);         
                if (methodBody != null) 
                {
                    processMethod(methodBase, methodBody);
                }
            }
        }

        internal void load(String[] args)
        {
            if (args.Length == 0)
            {
                Console.Error.WriteLine("Usage: cslint (option)* (assemly-path)*");
                Console.Error.WriteLine("Options:");
                Console.Error.WriteLine("\t-all-globals-reentrant:\ttreateall all global method as reentrant");
                Console.Error.WriteLine("\t-report-all-loops:\treport all detected loops");
                return;
            }
            AssemblyLoader loader = new AssemblyLoader();
            foreach (String arg in args) 
            {
                if (arg.Equals("-all-globals-reentrant"))
                {
                    allGlobalsReentrant = true;
                } 
                else if (arg.Equals("-report-all-loops")) 
                {
                    reportAllLoops = true;
                } 
                else if (arg.StartsWith("-")) 
                {
                    Console.Error.WriteLine("Urecognized option " + arg);
                    Console.Error.WriteLine("Supported options:");
                    Console.Error.WriteLine("\t-all-globals-reentrant");
                    Console.Error.WriteLine("\t-report-all-loops");
                } 
                else 
                {
                    Assembly assembly = Assembly.LoadFrom(arg);
                    foreach (Module module in assembly.GetModules()) 
                    {
                        reader = new ModuleReader(module, loader);
                        modules.Add(module);
                        foreach (Type type in module.GetTypes()) 
                        {
                            foreach (MemberInfo mbr in type.GetMembers(BindingFlags.Public | BindingFlags.NonPublic | BindingFlags.Static | BindingFlags.Instance | BindingFlags.DeclaredOnly))
                            {
                                MethodBase mb = mbr as MethodBase;
                                if (mb != null) 
                                {
                                    loadMethodBody(mb);
                                } 
                                else 
                                { 
                                    PropertyInfo pi = mbr as PropertyInfo;
                                    if (pi != null) 
                                    { 
                                        loadMethodBody(pi.GetGetMethod(true));
                                        loadMethodBody(pi.GetSetMethod(true));
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }


        internal MethodDesc getMethodDesc(MethodBase mb)
        {
            Type cls = mb.DeclaringType;
            return (cls != null) ? getClassDesc(mb.DeclaringType).getMethod(mb) : null;
        }

        internal FieldDesc getFieldDesc(FieldInfo fi)
        {
            return getClassDesc(fi.DeclaringType).getField(fi);
        }


        internal ClassDesc getClassDesc(Type type)
        {
            ClassDesc cls = (ClassDesc)classes[type];
            if (cls == null) 
            { 
                cls = new ClassDesc(type);
                classes[type] = cls;
            }
            return cls;
        }

        struct StackValue 
        {
            internal enum ValueType 
            {
                THIS,
                FIELD,
                METHOD,
                CALL
            };
            internal ValueType type;
            internal Object    link;

            internal void setField(Object fi) 
            {
                type = ValueType.FIELD;
                link = fi;
            }

            internal void setCall(Object md) 
            {
                type = ValueType.CALL;
                link = md;
            }

            internal void setMethod(Object mb)
            {
                type = ValueType.METHOD;
                link = mb;
            }
        }

        internal void processMethod(MethodBase methodBase, MethodBody methodBody)
        {
            ClassDesc selfClass = getClassDesc(methodBase.DeclaringType);
            MethodDesc methodDesc = selfClass.getMethod(methodBase);
            StackValue[] stack = new StackValue[methodBody.MaxStack < 16 ? 16 : methodBody.MaxStack+1];
            int sp = 0;

            if (!methodBase.IsAbstract)
            {
                MethodInfo mi = methodBase as MethodInfo;
                if (mi != null) 
                { 
                    MethodBase overriddenMethod = mi.GetBaseDefinition();
                    if (overriddenMethod != null && !overriddenMethod.Equals(mi)) 
                    {
                        MethodDesc md = getMethodDesc(overriddenMethod);
                        if (md != null) 
                        { 
                            md.addOverriddenBy(methodDesc);
                        }
                    }
                }
            }

            monitors.clear();
            topLevelSyncBlocks.clear();
            syncInsns.Clear();
	
            foreach (MethodImplAttribute attrib in methodBase.GetCustomAttributes(typeof(MethodImplAttribute), false))
            {
                if (attrib.Value == MethodImplOptions.Synchronized) 
                { 
                    GraphVertex monitor = methodBase.IsStatic 
                        ? selfClass.metaclassMonitor 
                        : selfClass.classMonitor;
                    monitors.push(monitor);
                    topLevelSyncBlocks.push(monitor);
                    if (methodBody.Instructions.Length > 0) 
                    {
                        syncInsns.Add(methodBody.Instructions[0]);
                    }
                    break;
                }
            }
            foreach (Instruction instruction in methodBody.Instructions)
            {
                OpCode opCode = instruction.OpCode;
                Object operand = instruction.Operand;
                Byte[] operandData = instruction.OperandData;
	
                if (opCode.Equals(OpCodes.Call)) 
                {
                    GraphVertex monitor;
                    if (operand.Equals(enter)) 
                    { 
                        if (sp > 0)
                        {
                            sp -= 1;
                            switch (stack[sp].type) 
                            {
                                case StackValue.ValueType.FIELD: 
                                {
                                    FieldDesc fd = getFieldDesc((FieldInfo)stack[sp].link);
                                    if (fd.syncObject == null) 
                                    {
                                        fd.syncObject = monitor = new GraphVertex();
                                        monitor.syncObject = fd;
                                    } 
                                    else 
                                    {
                                        monitor = fd.syncObject;
                                    }
                                    break;
                                }
                                case StackValue.ValueType.CALL: 
                                {
                                    MethodDesc md = (MethodDesc)stack[sp].link;
                                    if (md.syncObject == null) 
                                    {
                                        md.syncObject = monitor = new GraphVertex();
                                        monitor.syncObject = md;
                                    } 
                                    else 
                                    {
                                        monitor = md.syncObject;
                                    }
                                    break;
                                }
                                 case StackValue.ValueType.THIS:
                                    monitor = selfClass.classMonitor;
                                    break;
                                default:
                                    monitor = new GraphVertex();
                                    break;
                            }
                        } 
                        else 
                        {
                            monitor = new GraphVertex();
                        }
                        if (!monitors.empty()) 
                        { 
                            monitors.top().addEdge(monitor, methodDesc, instruction, CallContext.monitorEnter);
                        } 
                        else 
                        { 
                            syncInsns.Add(instruction);
                            topLevelSyncBlocks.push(monitor);
                        }
                        monitors.push(monitor);
                    } 
                    else if (operand.Equals(exit)) 
                    {
                        monitors.pop();
                    }
                }
                if (opCode.Equals(OpCodes.Call) || opCode.Equals(OpCodes.Calli) || opCode.Equals(OpCodes.Callvirt))
                {
                    MethodDesc callee = getMethodDesc((MethodBase)operand);
                    if (callee != null)
                    {
                        methodDesc.addInvocation(methodDesc, 
                            callee, 
                            opCode.Equals(OpCodes.Callvirt), 
                            instruction, 
                            monitors);
                        stack[0].setCall(callee);
                        sp = 1;
                    } 
                    else 
                    {
                        sp = 0;
                    }
                } 
                else if (opCode.Equals(OpCodes.Newobj)) 
                {
                    ConstructorInfo ci = operand as ConstructorInfo;
                    if (sp != 0 && stack[sp-1].type == StackValue.ValueType.METHOD 
                        && ci != null && ci.DeclaringType.Equals(typeof(ThreadStart)))
                    {
                        MethodDesc run = getMethodDesc((MethodBase)stack[--sp].link);
                        if (run != null) 
                        { 
                            run.nextEntryPoint = entryPoints;
                            entryPoints = run;
                        }
                    } 
                    else 
                    {
                        sp = 0;
                    }
                }
                else if (opCode.Equals(OpCodes.Ldfld) && operand is FieldInfo) 
                {
                    if (sp == 0) 
                    {
                        sp += 1;
                    }
                    stack[sp-1].setField(operand);
                } 
                else if (opCode.Equals(OpCodes.Ldftn) && operand is MethodBase) 
                {
                    stack[sp++].setMethod(operand);
                } 
                else if (opCode.Equals(OpCodes.Ldarg_0) && !methodBase.IsStatic) 
                {
                    stack[sp++].type = StackValue.ValueType.THIS;
                } 
                else if (opCode.Equals(OpCodes.Dup) && sp > 0) 
                {
                    stack[sp] = stack[sp-1];
                    sp += 1;
                }
                else if (opCode.Equals(OpCodes.Dup) && sp > 0) 
                {
                    stack[sp] = stack[sp-1];
                    sp += 1;
                }
                else if (sp > 0 
                    && opCode.StackBehaviourPush == StackBehaviour.Push0 
                    && (opCode.StackBehaviourPop == StackBehaviour.Popref 
                        || opCode.StackBehaviourPop == StackBehaviour.Pop1))
                {
                    sp -= 1;
                } 
                else 
                {
                    sp = 0;
                }
            }
            methodDesc.topLevelSyncBlocks = topLevelSyncBlocks.toArray();
            methodDesc.syncInsns = (Instruction[])syncInsns.ToArray(typeof(Instruction));
        }

        private sealed class AssemblyLoader : IAssemblyLoader
        {
            public Assembly Load(string assemblyName)
            {
                return Assembly.Load(assemblyName);	
            }
        }

        public virtual void message(String msg)
        {
            Console.Out.WriteLine(msg);
        }

        internal MethodInfo    enter;
        internal MethodInfo    exit;
        internal Hashtable     classes;
        internal ArrayList     syncInsns;
        internal VertexList    monitors;
        internal VertexList    topLevelSyncBlocks;
        internal bool          allGlobalsReentrant;
        internal bool          reportAllLoops;
        internal MethodDesc    entryPoints;
        internal ArrayList     modules;
        internal ModuleReader  reader;
        internal static CSLint instance = new CSLint();

        internal CSLint() 
        { 
            enter = typeof(System.Threading.Monitor).GetMethod("Enter");
            exit = typeof(System.Threading.Monitor).GetMethod("Exit");    
            classes = new Hashtable();
            syncInsns = new ArrayList();
            modules = new ArrayList();
            monitors = new VertexList();
            topLevelSyncBlocks = new VertexList();
            entryPoints = null;
            allGlobalsReentrant = false;
            reportAllLoops = false;
        }
    }
}
